/**
 * @file print.h Interface to Input event device info
 *
 * Copyright (C) 2010 Alfred E. Heggestad
 */


void print_name(int fd);
void print_events(int fd);
void print_keys(int fd);
void print_leds(int fd);
